use std::net::IpAddr;

use rocket::{
	form::{self, Contextual, Error, Form},
	response::Redirect,
	tokio::task,
};
use rocket_dyn_templates::{context, Template};

use crate::{
	auth::pw,
	db::{
		otp,
		schema::{DollTagsDb, User},
		user,
	},
	pages::CommonTemplateState,
	routes::error_handlers::PageResult,
};

#[get("/settings")]
pub async fn show_settings(
	mut db: DollTagsDb,
	user: User,
	meta: CommonTemplateState,
) -> PageResult {
	let has_otp = otp::has_otp(&mut *db, &user.id).await?;

	Ok(Template::render(
		"account/settings",
		context! {
			user,
			meta,
			otp: has_otp,
			prev_common: form::Context::default(),
			prev_password: form::Context::default(),
		},
	)
	.into())
}

#[derive(FromForm)]
pub struct CommonSettings<'a> {
	#[field(validate=len(..=256))]
	pub username: &'a str,
	#[field(validate=len(..=254))]
	pub email: &'a str,
}

#[post("/settings/common", data = "<form>")]
pub async fn change_settings(
	mut db: DollTagsDb,
	user: User,
	meta: CommonTemplateState,
	mut form: Form<Contextual<'_, CommonSettings<'_>>>,
	client_ip: IpAddr,
) -> PageResult {
	let values = match form.value {
		Some(ref v) => v,
		None => {
			return Ok(Template::render(
				"account/settings",
				context! {
					user,
					meta,
					prev_common: &form.context,
					prev_password: form::Context::default(),
				},
			)
			.into())
		}
	};

	if user.username != values.username && (user::exists(&mut *db, values.username).await?) {
		form.context
			.push_error(Error::validation("username already taken").with_name("username"));
		return Ok(Template::render(
			"account/settings",
			context! {
				user,
				meta,
				prev_common: &form.context,
				prev_password: form::Context::default(),
			},
		)
		.into());
	}

	let new_username = if values.username.len() != 0 {
		values.username
	} else {
		&user.username
	};
	// TODO: this will not allow to unset the email
	let new_email = if values.email.len() != 0 {
		Some(values.email)
	} else {
		user.email.as_ref().map(|w| w.as_str())
	};

	user::update_info(&mut *db, &user.id, new_username, new_email).await?;

	warn!(
		"[audit|{}] [{}] changed username/email (from {})",
		client_ip,
		user.id.to_string(),
		user.username,
	);

	Ok(Redirect::to(uri!("/account", show_settings)).into())
}

#[derive(FromForm)]
pub struct PasswordSettings<'a> {
	#[field(validate=len(8..))]
	pub old_password: &'a str,
	#[field(validate=len(8..))]
	#[field(validate=validate_new_passwords(self.confirm_password))]
	pub password: &'a str,
	#[field(validate=len(8..))]
	pub confirm_password: &'a str,
}

pub fn validate_new_passwords<'v>(new: &str, confirm: &str) -> form::Result<'v, ()> {
	if new != confirm {
		Err(form::Error::validation("the two new passwords don't match"))?;
	}

	Ok(())
}

#[post("/settings/password", data = "<form>")]
pub async fn change_password(
	mut db: DollTagsDb,
	user: User,
	meta: CommonTemplateState,
	mut form: Form<Contextual<'_, PasswordSettings<'_>>>,
	client_ip: IpAddr,
) -> PageResult {
	let values = match form.value {
		Some(ref v) => v,
		None => {
			return Ok(Template::render(
				"account/settings",
				context! {
					user,
					meta,
					prev_common: form::Context::default(),
					prev_password: &form.context,
				},
			)
			.into())
		}
	};

	let password = String::from(values.old_password);
	let old_password = user.password.clone();
	let right_password =
		task::spawn_blocking(move || pw::verify(&password, &old_password)).await??;
	if !right_password {
		form.context
			.push_error(Error::validation("wrong password").with_name("old_password"));
		return Ok(Template::render(
			"account/settings",
			context! {
				user,
				meta,
				prev_common: form::Context::default(),
				prev_password: &form.context,
			},
		)
		.into());
	}

	let new_password = String::from(values.password);
	let new_password_hash = task::spawn_blocking(move || pw::hash(&new_password)).await??;
	user::update_password(&mut *db, &user.id, &new_password_hash).await?;

	warn!(
		"[audit|{}] [{}] changed password",
		client_ip,
		user.id.to_string()
	);

	Ok(Redirect::to(uri!("/account", show_settings)).into())
}
